/*
 * This file is part of battwd. See battwd.c for further information.
 *
 * battwd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * battwd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with battwd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include "config.h"
#include "global.h"

/* Private variables */
static volatile int interrupted;
static char buffer[BUF_SIZE];

/* Private functions */
static int to_int(char *str);
static void sigchld(int signum);
static void init(battery *battp);
static void down(battery *battp);
static int run_control(battery *battp);
static int exec_action(const action *act);
static int read_power_status(battery *battp);
static void ptmsg(int level, const char *format, ...);
static int get_param(const mon_file *pwfiles, int line);
static void set_power_value(power_status *pws, int param, char *val);
static void parse_file(FILE *bsfp, const mon_file *file, power_status *pws);

/**
 *
 */
static void ptmsg(int level, const char *format, ...)
{
  va_list ap;
  extern volatile int daemon_flag;
 
  va_start(ap, format);
  
  if (!daemon_flag) {
    vfprintf(stderr, format, ap);
    fprintf(stderr, "\n");
  } else {
    vsyslog(level, format, ap);
  }
}

/**
 *
 */
static int to_int(char *str)
{
  int i;
  extern char buffer[BUF_SIZE];

  for (i = 0; i < strlen(str); i++) {
    if (!isdigit(*(str + i)))
      break;
  }

  return atoi(strncpy(buffer, str, i));
}

/**
 *
 */
static int exec_action(const action *act)
{
  pid_t child_pid, tpid;
  int child_status;
  
  child_pid = fork();
  
  if(child_pid == 0) {
    /* This is done by the child process. */
    execv(act->path, (char * const*)act->argv);
    
    /* If execv returns, it must have failed. */
    ptmsg(LOG_ERR,
          "Error executing action script: %s %s",
          act->path,
          strerror(errno));
    exit(EXIT_FAILURE);
  } else {
    /* This is run by the parent.  Wait for the child
       to terminate. */
    do {
      /* Success = ID of the terminated child;
         Error   = -1 */
      tpid = wait(&child_status);
      if(WIFEXITED(child_status) && 
         (WEXITSTATUS(child_status) == EXIT_FAILURE))
        return ERR_EXEC_ACTION;
    } while(tpid != child_pid);
  }
  return tpid;
}

/**
 *
 */
static int get_param(const mon_file *pwfiles, int line)
{
  int i;

  for (i = 0; i < LENGTH(pwfiles->alist); i++)
    if (((pwfiles->alist) + i)->nline == line)
      return ((pwfiles->alist) + i)->param;
  return -1;
}

/**
 *
 */
static void set_power_value(power_status *pws, int param, char *val)
{
  switch(param) {
  case BAT_PRESENT:
    strcpy(pws->bat_present, val);
    break;
  case BAT_CAPACITY:
    strcpy(pws->bat_capacity, val);
    break;
  case BAT_CHARGING:
    strcpy(pws->bat_charging, val);
    break;
  case BAT_RATE:
    pws->bat_rate = to_int(val);
    break;
  case BAT_REMAINING:
    pws->bat_remaining = to_int(val);
    break;
  case BAT_VOLTAGE:
    pws->bat_voltage = to_int(val);
    break;
  case BAT_ALARM:
    pws->bat_alarm = to_int(val);
    break;
  case AC_ADAPTER:
    strcpy(pws->ac_adapter, val);
    break;
  }  
}

/**
 *
 */
static void parse_file(FILE *bsfp, const mon_file *file, power_status *pws)
{
  int i, c, nval, lc;
  extern char buffer[BUF_SIZE];

  i = 0;
  lc = 0;
  
  if (file->delimiter)
    nval = 1;
  else
    nval = 0;
  
  while ((c = getc(bsfp)) != EOF) {
    if (nval && c != DELIMITER_0 && c != DELIMITER_1)
      continue;
    nval = 0;
    if (c == '\n' || c == '\r') {
      buffer[i] = '\0';
      i = 0;
      set_power_value(pws, get_param(file, lc), buffer);
      nval = ++lc;
      continue;
    }
    if (c != ' ' && c != '\t' && c != DELIMITER_0 && c != DELIMITER_1)
      buffer[i++] = c;
  }
}

/**
 *
 */
static int read_power_status(battery *battp)
{
  int i;
  FILE *fp;
  
  /* Read state file files */
  for (i = 0; i < battp->num_files; i++) {
    if ((fp = fopen (battp->files[i].path, "r")) != NULL) {
      parse_file(fp, &battp->files[i], battp->status);
    } else {
      return(ERR_READING_STATE);
    }
    if (fclose(fp) != 0)
      return(ERR_CLOSING_STATE);
  }
  
  return i;
}

/**
 *
 */
static int run_control(battery *battp)
{
  int i;
  extern volatile int interrupted;
  
  interrupted = 0;
  while(!interrupted) {
    if (read_power_status(battp) == ERR_READING_STATE) {
      ptmsg(LOG_ERR, "Sorry, I can't read information for battery %s",
            battp->id);
      return ERR_READING_STATE;
    }
    for (i = 0; i < battp->num_actions; i++) {
      /* Checking the battery charge */
      if ((battp->status->bat_alarm * battp->actions[i].level) 
          >= battp->status->bat_remaining) {
        /* Cheking the AC status */
        if (strcmp(battp->status->ac_adapter, ON_LINE_0) != 0 &&
            strcmp(battp->status->ac_adapter, ON_LINE_1) != 0) {
          /* Cheking action flags */
          if (!*(battp->action_flags + i)) {
            ptmsg(LOG_INFO, "[%s] Executing action %s",
                  battp->id,
                  battp->actions[i].path);
            /* Execute an action */
            if (exec_action(&battp->actions[i]) == ERR_EXEC_ACTION) {
              ptmsg(LOG_INFO, "[%s] An error occurred excuting action %s",
                    battp->id,
                    battp->actions[i].path);
            } else {
              *(battp->action_flags + i) = 1;
            }
          }
        } else {
          if (*(battp->action_flags + i))
            *(battp->action_flags + i) = 1;
        }
      }
    }
    sleep(SLEEP_SECONDS);
  }
  return i;
}

/**
 *
 */
static void init(battery *battp)
{
  battp->status = malloc(sizeof(power_status));
  battp->status->ac_adapter = malloc(sizeof(char)*BUF_SIZE);
  battp->status->bat_present = malloc(sizeof(char)*BUF_SIZE);
  battp->status->bat_capacity = malloc(sizeof(char)*BUF_SIZE);
  battp->status->bat_charging = malloc(sizeof(char)*BUF_SIZE);
  battp->action_flags = malloc(sizeof(int)*(battp->num_actions));
}

/**
 *
 */
static void down(battery *battp)
{
  free((void *)battp->action_flags);
  free(battp->status->bat_charging);
  free(battp->status->bat_capacity);  
  free(battp->status->bat_present);
  free(battp->status->ac_adapter);
  free(battp->status);
}

/**
 *
 */
static void sigchld(int signum)
{
  extern volatile int interrupted;

  switch(signum) {
  case SIGHUP:
    interrupted = signum;
    break;
  case SIGINT:
    interrupted = signum;
    break;
  case SIGTERM:
    interrupted = signum;
    break;
  }
}

/**
 *
 */
extern int run_daemon()
{
  /* Our process ID and Session ID */
  pid_t pid, sid;
  struct sigaction *sighandler;
  extern volatile int daemon_flag;
  extern volatile int interrupted;
  extern battery batteries[];

  battery *batts = &batteries[0];

  sighandler = malloc(sizeof(sighandler));
  sighandler->sa_handler = sigchld;
 
  if (sigaction(SIGHUP, sighandler, NULL) ||
      sigaction(SIGINT, sighandler, NULL) ||
      sigaction(SIGTERM, sighandler, NULL)) {
    perror("Error registering signal actions.");
    return EXIT_FAILURE;
  }
  
  openlog("battwd", LOG_CONS, LOG_USER);
  ptmsg(LOG_INFO, "Battery watcher daemon " VERSION " started");

  if (daemon_flag) {
    /* Fork off the parent process */
    pid = fork();
    if (pid < 0) {
      ptmsg(LOG_ERR, "Error creating child process: %s", strerror(errno));
      return ERR_CHILD_PROC;
    }
    /* If we got a good PID, then
       we can exit the parent process. */
    if (pid > 0) {
      return EXIT_SUCCESS;
    }
    
    /* Change the file mode mask */
    umask(0);
    
    /* Create a new SID for the child process */
    sid = setsid();
    if (sid < 0) {
      ptmsg(LOG_ERR, "Error creating new SID for child process: %s",
            strerror(errno));
      return ERR_CHILD_SID;
    }
    
    /* Change the current working directory */
    if ((chdir("/")) < 0) {
      ptmsg(LOG_ERR, "Error changing directory: %s",
            strerror(errno));
      return ERR_CHDIR;
    }
    
    /* Close out the standard file descriptors */
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
  }
  
  /* Initialization */
  init(batts + 0);
  init(batts + 1);

  /* read_power_status(batts + 0); */
  if (read_power_status(batts + 1) == ERR_READING_STATE) {
    ptmsg(LOG_ERR, "Sorry, I can't read information for battery %s",
          (batts + 1)->id);
    return ERR_READING_STATE;
  }
  
  /* Log initial estate */
  ptmsg(LOG_INFO, "[%s] Status: alarm %d; remaining %d; voltage %d",
        (batts + 1)->id,
        (batts + 1)->status->bat_alarm,
        (batts + 1)->status->bat_remaining,
        (batts + 1)->status->bat_voltage);
  ptmsg(LOG_INFO, "[%s] AC adapter state: %s",
        (batts + 1)->id,
        (batts + 1)->status->ac_adapter);

  /** 
   * Control loop 
   **/
  run_control(batts + 1);

  ptmsg(LOG_INFO, "Interrupted by %s signal", strsignal(interrupted));
  
  closelog();
  down(batts + 1);
  /* down(batts + 0); */
  free(sighandler);
  
  return INT_BY_SIGNAL;
}
